"""
Functions which can be used to manipulate data items

Available to the user in the command terminal

"""

from Qt.QtWidgets import (QDialog, QGridLayout, QDialogButtonBox,
                          QPushButton, QInputDialog)
from Qt.QtCore import Qt

import numpy as np
from pyxpad import calculus
from .pyxpad_utils import XPadDataItem, XPadDataDim


def XPadFunction(func, name="f"):
    """
    Turns a NumPy function into a function of data item
    """
    def newfunc(data):
        result = XPadDataItem()
        if data.name != "":
            result.name = name + "( " + data.name + " )"
        result.source = data.source
        if data.label != "":
            result.label = name + "( " + data.label + " )"
        result.data   = func(data.data)
        result.dim    = data.dim
        result.order  = data.order
        result.time   = data.time
        return result
    return newfunc


sin = XPadFunction(np.sin, "sin")
cos = XPadFunction(np.cos, "cos")
tan = XPadFunction(np.tan, "tan")

exp = XPadFunction(np.exp, "exp")
log = XPadFunction(np.log, "log")

sqrt = XPadFunction(np.sqrt, "sqrt")

def reciprocal(data):
    recip = XPadFunction(np.reciprocal, "recip")
    reciprocal = recip(data)
    reciprocal.units = data.units + chr(0x207B) + chr(0x00B9)
    return reciprocal

def exponential(data):
    exp = XPadFunction(np.exp, "exp")
    return exp(data)

def absolute(data):
    abso = XPadFunction(np.absolute, "abs")
    return abso(data)

def arctan(data):
    atan = XPadFunction(np.arctan, "arctan")
    return atan(data)

def nlog(data):
    natlog = XPadFunction(np.log, "nlog")
    return natlog(data)

def normalise(data):
    point = len(data.data)-1
    integral = calculus.integrate(data)
    normfac = integral.data[point]
    normdat = np.true_divide(data.data, normfac)
    result = XPadDataItem(integral)
    result.data[:] = normdat[:]
    if data.name != "":
        result.name = "Norm(" + data.name + ")"
    if data.label != "":
        result.label = "Norm(" + data.label + ")"
    result.units = data.units
    return result

def invert(data):
    result = XPadDataItem(data)
    result.data = -1.*data.data
    result.name = "-(" + data.name + ")"
    result.label = "-(" + data.label + ")"
    return result

def addcon(data):
    result = XPadDataItem(data)
    parent = QDialog()
    dialog = QInputDialog.getDouble(parent, "X+C", "Input constant:")
    if dialog[1]:
        constant = dialog[0]
    result.data = np.add(data.data, constant)
    result.name = "(" + data.name + ")+" + str(constant)
    result.label = "(" + data.label + ")+" + str(constant)
    return result

def subcon(data):
    result = XPadDataItem(data)
    parent = QDialog()
    dialog = QInputDialog.getDouble(parent, "X-C", "Input constant:")
    if dialog[1]:
        constant = dialog[0]
    constant = 1.
    result.data = np.subtract(data.data, constant)
    result.name = "(" + data.name + ")-" + str(constant)
    result.label = "(" + data.label + ")-" + str(constant)
    return result

def mulcon(data):
    result = XPadDataItem(data)
    parent = QDialog()
    dialog = QInputDialog.getDouble(parent, "X*C", "Input constant:")
    if dialog[1]:
        constant = dialog[0]
    result.data = np.multiply(data.data, constant)
    result.name = "(" + data.name + ")*" + str(constant)
    result.label = "(" + data.label + ")*" + str(constant)
    return result

def divcon(data):
    result = XPadDataItem(data)
    parent = QDialog()
    dialog = QInputDialog.getDouble(parent, "X/C", "Input constant:")
    if dialog[1]:
        constant = dialog[0]
    result.data = np.true_divide(data.data, constant)
    result.name = "(" + data.name + ")/" + str(constant)
    result.label = "(" + data.label + ")/" + str(constant)
    return result

def powcon(data):
    result = XPadDataItem(data)
    parent = QDialog()
    dialog = QInputDialog.getDouble(parent, "X^C", "Input constant:")
    if dialog[1]:
        constant = dialog[0]
    result.data = np.power(data.data, constant)
    result.name = "(" + data.name + ")^" + str(constant)
    result.label = "(" + data.label + ")^" + str(constant)
    result.units = "(" + data.units + ")^" + str(constant)
    return result

def inputname():
    parent = QDialog()
    title = "Change Name"
    label = "Enter new name:"
    dialog = QInputDialog.getText(parent, title, label)
    if dialog[1]:
        name = dialog[0]
    return name

def changename(data):
    result = XPadDataItem(data)
    result.name = "Renamed(" + data.name +")"
    return result

def changeunits(data):
    result = XPadDataItem(data)
    parent = QDialog()
    title = "Change Units"
    label = "Enter new units:"
    dialog = QInputDialog.getText(parent, title, label)
    if dialog[1]:
        newUnit = dialog[0]
        unitFact = getUnitFactor()
    if unitFact and newUnit:
        result.data = np.multiply(data.data, unitFact)
        result.units = newUnit
        return result

def getUnitFactor():
    parent = QDialog()
    title = "Change Units"
    label = "Enter conversion factor:"
    dialog = QInputDialog.getDouble(parent, title, label)
    if dialog[1]:
        return dialog[0]
    else:
        return None

def statistics(data):
    n = len(data.data)
    sumOfData = np.sum(data.data)
    mean = sumOfData/n
    dataSquared = np.power(data.data, 2)
    variance = (np.sum(dataSquared)/n)-mean**2.
    stdDev = np.sqrt(variance)
    minVal = data.data[0]
    maxVal = data.data[0]
    for value in data.data:
        if value >= maxVal:
            maxVal = value
        if value <= minVal:
            minVal = value
    rng = "(" + str(minVal) +")" + " - " + "(" + str(maxVal) + ")"
    stats = "\nStatistics on " + data.name + ":"
    stats += "\nMean = " + str(mean) + "\nStandard Deviation = " + str(stdDev)
    stats += "\nRange = " + rng
    print(stats)
    return()

def timeOffset(data):
    result = XPadDataItem(data)
    parent = QDialog()
    title = "Time Offset"
    label = "Input time offest:"
    dialog = QInputDialog.getDouble(parent, title, label)
    if dialog[1]:
        offset = dialog[0]
        result.dim[data.order].data = np.add(result.dim[data.order].data, offset)
        result.name = "Timoff(" + result.name + ", " + str(offset) + ")"
        result.label = "Timoff(" + result.label + ", " + str(offset) + ")"
        return result

def chop(item, t_min, t_max):
    """
        >>> from user_functions import *
        >>> a = chop(XMC_OMV_110, 0.274, 0.276)
        >>> a_amp,a_phase = fftp(a)
        >>> b = chop(a_phase, 0.0, 100.0)
        >>> plot(b)
    """
    if len(item.dim) != 1:
        raise ValueError("chop can only operate on 1D traces currently")

    if t_max < t_min or t_max < item.dim[0].data[0] or t_min > item.dim[0].data[-1]:
        raise ValueError("New time-range not defined correctly")

    idx = np.where(np.logical_and(item.dim[0].data >= t_min, item.dim[0].data <= t_max))

    if len(idx[0]) == 0:
        raise ValueError("No data in time-range specified")

    # Calculate the phase
    chopped = XPadDataItem()

    if item.name != "":
        chopped.name = "CHOP( "+item.name+", "+str(t_min)+", "+str(t_max)+" )"
    chopped.source = item.source
    if item.label != "":
        chopped.label = "CHOP( "+item.label+", "+str(t_min)+", "+str(t_max)+" )"
    chopped.units = item.units

    chopped.data = item.data[idx]

    # Create a dimension
    dim = XPadDataDim()

    dim.name = item.dim[0].name
    dim.label = item.dim[0].label
    dim.units = item.dim[0].units

    dim.data = item.dim[0].data[idx]
    chopped.dim = [dim]

    if chopped.dim[0].units in ["s", "S", "sec", "Sec", "SEC"]:
        chopped.order = 0
        chopped.time = chopped.dim[0].data

    return chopped

def clip(item, valmin, valmax):
    """
    Removes values outside a given range
    """

    if len(item.dim) != 1:
        raise ValueError("Clip can only operate on 1D traces currently")

    if valmax < valmin:
        raise ValueError("Clip range incorrectly defined")

    clipped = XPadDataItem(item)
    data = []
    time = []

    for point in range(len(item.data)-1):
        if item.data[point] <= valmax and item.data[point] >= valmin:
            data += [item.data[point]]
            time += [item.dim[item.order].data[point]]

    if len(data) == 0:
        raise ValueError("No data in the specified range")

    clipped.data = data
    clipped.dim[clipped.order].data = time
    clipped.time = time
    clipped.name = "CLIP("+item.name+", "+str(valmin)+", "+str(valmax)+")"
    clipped.label = "CLIP("+item.label+", "+str(valmin)+", "+str(valmax)+")"

    return clipped

##########################################################################

"""
    from user_functions import *
    data = read_padsav("12354-omv110-FFTP.padsav")[0]
    plot(data)
"""

from numpy import linspace, rollaxis
from scipy.io import readsav
from warnings import catch_warnings, simplefilter


def read_padsav(file_name, disable_UserWarnings=True):
    """
        Reads data from an idl_dict object into XPadDataItem objects

        Parameters
        ----------
        file_name : str
            Path to XPad*.padsav file


        Returns
        -------
        items : list
            A list of XPadDataItems
    """

    warning_action = "default"
    if (disable_UserWarnings):
        warning_action = "ignore"

    with catch_warnings():
        simplefilter(warning_action, UserWarning)
        idl_dict = readsav(file_name)

    return parse_padsav(idl_dict)


def parse_padsav(xpad_idl_dict):
    """
        Reads data from an idl_dict object into XPadDataItem objects

        Parameters
        ----------
        xpad_idl_dict : AttrDict or dict
            Input dict from an XPad *.padsav imported using scipy.io.readsav


        Returns
        -------
        items : list
            A list of XPadDataItems
    """

    if not check_padsav(xpad_idl_dict):
        return None

    items = []
    for trace in xpad_idl_dict['ptr']:
        item = XPadDataItem()

        item.name = trace['NAME'][0]
        item.source = trace['SOURCE'][0]
        item.label = trace['DINFO'][0]['LABEL'][0]
        item.units = trace['DINFO'][0]['UNITS'][0]

        numdims = trace['SIZE'][0][0]

        item.data = rollaxis(trace['DATA'][0], numdims - 1)

        if numdims > 0:
            # f(t), f(t,x) or f(t,x,y)
            dim = XPadDataDim()

            # If we have a time domain, use data from TINFO
            if trace['TINFO'][0]['DOMAINS'][0] > 0:
                dim.data = linspace(trace['TINFO'][0]['START'][0],
                                    trace['TINFO'][0]['FINISH'][0],
                                    trace['TINFO'][0]['LENGTH'])
            else:
                dim.data = trace['TIME'][0]

            dim.name = trace['TINFO'][0]['LABEL'][0]
            dim.label = dim.name
            dim.units = trace['TINFO'][0]['UNITS'][0]

            item.dim.append(dim)

            item.order = len(item.dim) - 1
            item.time = item.dim[item.order].data

        if numdims > 2:
            # f(t,x,y)
            dim = XPadDataDim()
            dim.data = trace['Y'][0]
            dim.name = trace['YINFO'][0]['LABEL'][0]
            dim.label = dim.name
            dim.units = trace['YINFO'][0]['UNITS'][0]

            item.dim.append(dim)

        if numdims > 1:
            # f(t,x) or f(t,x,y)
            dim = XPadDataDim()
            dim.data = trace['X'][0]
            dim.name = trace['XINFO'][0]['LABEL'][0]
            dim.label = dim.name
            dim.units = trace['XINFO'][0]['UNITS'][0]

            item.dim.append(dim)

        item.desc = trace['TYPE'][0]

        items.append(item)

    return items


def check_padsav(xpad_idl_dict):
    """
        Verifies padsav data and its containing traces are valid.

        Parameters
        ----------
        xpad_idl_dict : AttrDict or dict
            An XPad *.padsav imported using scipy.io.readsav


        Returns
        -------
        int : 0 for invalid, 1 for valid
    """
    if not isinstance(xpad_idl_dict, dict):
        return 0

    if 'ptr' not in xpad_idl_dict:
        return 0

    # [*] 'ptr not found in xpad_idl_dict'
    if not xpad_idl_dict['ptr'].size:
        return 0

    for trace in xpad_idl_dict['ptr']:
        # [*] failed on trace
        if not check_trace(trace):
            return 0

    return 1


def check_trace(xpad_idl_trace):
    """
        Verifies a trace is well-formed and valid

        Parameters
        ----------
        xpad_idl_trace : numpy.core.records.recarray
            A trace from imported XPad *.padsav data


        Returns
        -------
        int : 0 for invalid, 1 for valid
    """
    from numpy import prod
    from numpy.core.records import recarray

    # [*] 'xpad_idl_trace is not an instance of numpy.core.records.recarray'
    if not isinstance(xpad_idl_trace, recarray):
        return 0

    # [*] 'UTYPE != \'DBstructure\''
    if 'UTYPE' not in xpad_idl_trace.dtype.names or not xpad_idl_trace['UTYPE'][0] == 'DBstructure':
        return 0

    for field in ['TYPE', 'NAME', 'DATA', 'DINFO', 'SOURCE', 'PROCESS', 'SIZE', 'TINFO']:
        # [*] 'Required field missing: '+field
        if field not in xpad_idl_trace.dtype.names:
            return 0

    # [*] TINFO.UTYPE != \'TINFO\'
    if 'UTYPE' not in xpad_idl_trace['TINFO'][0].dtype.names or not xpad_idl_trace['TINFO'][0]['UTYPE'][0] == 'TINFO':
        return 0

    for field in ['DOMAINS', 'START', 'FINISH', 'STEP', 'SAMPLES', 'LENGTH', 'UNITS', 'LABEL']:
        # [*] 'Required TINFO field missing: '+field
        if field not in xpad_idl_trace['TINFO'][0].dtype.names:
            return 0

    types = ['f(t)', 'f(t,x)', 'f(t,x,y)']
    # [*] 'TYPE is not one of '+str(types)
    if not xpad_idl_trace['TYPE'][0] in types:
        return 0

    sizes = xpad_idl_trace['SIZE'][0]
    numdims = sizes[0]
    # [*] 'size(DATA) != prod(dim_sizes)'
    if not xpad_idl_trace['DATA'][0].size == prod(sizes[1:
    ]): return 0

    # [*] 'Incorrect numdims for shape of DATA'
    if not numdims == len(xpad_idl_trace['DATA'][0].shape):
        return 0

    # [*] 'Incorrect numdims in SIZE for TYPE'
    if not numdims == types.index(xpad_idl_trace['TYPE'][0]) + 1:
        return 0

    if numdims > 0:
        if xpad_idl_trace['TINFO'][0]['DOMAINS'][0] < 1:
            if 'TIME' in xpad_idl_trace.dtype.names:
                # [*] 'size(TIME) != SIZE[1]':
                if not xpad_idl_trace['TIME'][0].size == sizes[1]:
                    return 0

                # Should this always be true?
                # [*] 'size(TIME) != TINFO.LENGTH':
                #if not xpad_idl_trace['TIME'][0].size == xpad_idl_trace['TINFO'][0]['LENGTH'][0]: return 0
            else:
                # Can't determine TIME dimension
                # [*] 'TINFO contains no domains and TIME field does not exist'
                return 0
        else:
            # [*] 'TINFO.START > TINFO.FINISH'
            if not xpad_idl_trace['TINFO'][0]['START'][0] <= xpad_idl_trace['TINFO'][0]['FINISH'][0]:
                return 0

            # [*] 'TINFO.LENGTH is < 1'
            if not xpad_idl_trace['TINFO'][0]['LENGTH'][0] >= 1:
                return 0

    if numdims > 1:
        for field in ["X", "XINFO"]:
            # [*] 'Required field missing: '+field
            if field not in xpad_idl_trace.dtype.names:
                return 0

        # [*] 'size(X) != SIZE[2]':
        if not xpad_idl_trace['X'][0].size == sizes[2]:
            return 0

    if numdims > 2:
        for field in ["Y", "YINFO"]:
            # [*] 'Required field missing: '+field
            if field not in xpad_idl_trace.dtype.names:
                return 0

        # [*] 'size(Y) != SIZE[3]':
        if not xpad_idl_trace['Y'][0].size == sizes[3]:
            return 0

    # Add more checks if you so desire
    # ...
    return 1



#
# Initial code taken from  http://stackoverflow.com/questions/6723527/getting-pyside-to-work-with-matplotlib
# Additional bits from https://gist.github.com/jfburkhart/2423179
# zplot zoom effect from http://matplotlib.org/examples/pylab_examples/axes_zoom_effect.html
#

import matplotlib

from Qt import __qt_version__

# Use the correct backend for matplotlib, depending on Qt version and
# the particular Qt python bindings we're using
if __qt_version__.split('.')[0] == '5':
    matplotlib.use('Qt5Agg')
    from matplotlib.backends.backend_qt5agg import (FigureCanvasQTAgg as FigureCanvas,
                                                    NavigationToolbar2QT as NavigationToolbar)
else:
    matplotlib.use('Qt4Agg')
    from matplotlib.backends.backend_qt4agg import (FigureCanvasQTAgg as FigureCanvas,
                                                    NavigationToolbar2QT as NavigationToolbar)

from matplotlib.figure import Figure
from matplotlib.transforms import Bbox, TransformedBbox, \
    blended_transform_factory
from mpl_toolkits.axes_grid1.inset_locator import BboxPatch, BboxConnector,\
    BboxConnectorPatch

from matplotlib.pyplot import setp

from Qt.QtWidgets import *
from Qt.QtCore import Qt

import numpy as np


class MatplotlibWidget():

    def __init__(self, parent):

        self.figure = Figure()
        self.canvas = FigureCanvas(self.figure)
        self.canvas.setParent(parent)
        self.mpl_toolbar = NavigationToolbar(self.canvas, parent)
        self.axes = self.figure.add_subplot(111)

        self.grid_layout = QVBoxLayout()
        self.grid_layout.addWidget(self.canvas)
        self.grid_layout.addWidget(self.mpl_toolbar)
        parent.setLayout(self.grid_layout)

        self.callback_id = None

    def _clean_axes(self):
        """
        Make sure the figure is in a nice state
        """
        # Get rid of any extra axes
        if isinstance(self.axes, list):
            for axes in self.axes:
                del(axes)
            self.axes = self.figure.add_subplot(111)
        else:
            self.axes.clear()

        self.figure.clear()
        self.axes.grid(True)
        # Reset to some hardcoded default values
        self.figure.subplots_adjust(left=0.125, right=0.9, top=0.9, bottom=0.1,
                                    wspace=0.2, hspace=0.2)
        # Remove any event callbacks
        if self.callback_id:
            try:
                self.figure.canvas.mpl_disconnect(self.callback_id)
            except TypeError:
                for callback_id in self.callback_id:
                    self.figure.canvas.mpl_disconnect(callback_id)
                self.callback_id = None

    def plot(self, *args):
        """
        Make multiple plots
        """
        nplots = len(args)
        self._clean_axes()

        for plotnum, p in enumerate(args, 1):
            # For each plot
            if plotnum == 1:
                self.axes = self.figure.add_subplot(nplots, 1, plotnum)
                ax = self.axes
            else:
                self.axes = self.figure.add_subplot(nplots, 1, plotnum, sharex=ax)
            setp(self.axes.get_xticklabels(), visible=False)
            try:
                # Assume each p is a list of data items to be overplotted
                for data in p:
                    # Plot data
                    #self.axes.plot(data.time, data.data)
                    label = data.desc
                    if label == "":
                        label = data.label
                    label += " " + data.source

                    time = data.time
                    if time is None:
                        if len(data.dim) != 1:
                            print(data.dim)
                            raise ValueError("Cannot plot '"+data.label+"' as it has too many dimensions")
                        time = data.dim[0].data

                    self.axes.plot(time, data.data, label=label)

                # Y label from last plot
                ylabel = data.desc
                if ylabel == "":
                    ylabel = data.label
                    if data.units != "":
                        ylabel += " ("+data.units+")"
                self.axes.set_ylabel(ylabel)

                if plotnum == nplots:
                    self.axes.set_xlabel(p[0].dim[p[0].order].label)
                    setp(self.axes.get_xticklabels(), visible=True)
                self.axes.legend()
            except TypeError:
                # p not iterable, so just plot item
                #self.axes.plot(p.time, p.data)
                time = p.time.data
                xlabel = p.dim[p.order].label
                if time is None:
                    if len(p.dim) != 1:
                        raise ValueError("Cannot plot '"+p.label+"' as it has too many dimensions")
                    time = p.dim[0].data
                    xlabel = p.dim[0].label

                data = p.data
                # Check the size of the array
                size = len(time)
                if size > 10000:
                    fac = int(len(time) / 10000)
                    time = time[::fac]
                    data = data[::fac]
                    print("Warning: too many samples (%d). Down-sampling to %d points" % (size, len(time)))

                self.axes.plot(time, data)

                ylabel = p.desc
                if ylabel == "":
                    ylabel = p.label
                    if p.units != "":
                        ylabel += " ("+p.units+")"
                self.axes.set_ylabel(ylabel)
                if plotnum == 0:
                    self.axes.set_xlabel(xlabel)

        self.figure.subplots_adjust(left=0.08, right=0.98, top=0.95, bottom=0.07, hspace=0.001)

        self.canvas.draw()

    def plotxy(self, x, y):
        """
        Plot one variable against another
        """
        self.axes.clear()
        self.figure.clear()
        self.axes = self.figure.add_subplot(111)
        self.figure.subplots_adjust(left=0.07, right=0.98, top=0.95, bottom=0.08)
        self.axes.plot(x.data, y.data)
        self.axes.set_xlabel(x.name)
        self.axes.set_ylabel(y.name)
        self.canvas.draw()

    def oplot(self, *args):
        """
        Make an overplot from multiple traces
        """
        ntraces = len(args)
        if len(args) == 1:
            self.plot(*args)

        self._clean_axes()

        self.axes = self.figure.add_subplot(1, 1, 1)
        for tracenum, trace in enumerate(args):
            try:
                for data in trace:
                    label = data.desc
                    if label == "":
                        label = data.name + " (" + data.units + ") " + data.source

                    time = data.time
                    if time is None:
                        if len(data.dim) != 1:
                            print(data.dim)
                            raise ValueError("Cannot plot '"+data.label+"' as it has too many dimensions")
                        time = data.dim[0].data

                    self.axes.plot(time, data.data, label=label)
                if tracenum == 0:
                    ylabel = data.desc
                    if ylabel == "":
                        ylabel = data.label
                        if data.units != "":
                            ylabel += " ("+data.units+") "
                    self.axes.set_ylabel(ylabel)
                    self.axes.set_xlabel(trace[0].dim[trace[0].order].label)
                self.axes.legend()
            except TypeError:
                #Trace not iterable
                time = trace.time.data
                xlabel = trace.dim[trace.order].label
                if time is None:
                    if len(trace.dim) != 1:
                        raise ValueError("Cannot plot '"+trace.label+"' as it has too many dimensions")
                    time = trace.dim[0].data
                    xlabel = trace.dim[0].label

                data = trace.data
                # Check array size
                size = len(time)
                if size > 10000:
                    fac = int(len(time) / 10000)
                    time = time[::fac]
                    data = data[::fac]
                    print("Warning: too many samples (%d). Down-sampling to %d points" % (size, len(time)))

                self.axes.plot(time, data)

                ylabel = trace.desc
                if ylabel == "":
                    ylabel = trace.label
                    if trace.units != "":
                        ylabel += " ("+trace.units+") "
                self.axes.set_ylabel(ylabel)

        self.canvas.draw()

    def mplot(self, *args):
        """
        Make a custom number of plots with a custom number of traces on each
        """
        # Create a dialog box to get the plot format
        def getFormat():
            parent = QDialog()
            title = "Plot Format"
            label = 'Enter number of traces in each plot separated by ", ":'
            dialog = QInputDialog.getText(parent, title, label)
            if dialog[1]:
                form = dialog[0].split(',')
                form = [int(i) for i in form]
                return form
            else:
                raise ValueError("No format entered")

        pltform = getFormat()
        nplots = len(pltform)
        ntraces = len(args)

        # Check validity of the plot format
        if ntraces != np.sum(pltform):
            raise ValueError("Number of traces does not equal sum of traces to be plotted in format")

        self._clean_axes()
        # Split data items up into correct plots according to plot format
        traces = []
        maxindx = 0
        for i in range(nplots):
            plotdat = []
            minindx = maxindx
            maxindx += pltform[i]
            for j in range(ntraces):
                if j >= minindx and j < maxindx:
                    plotdat.append(args[j][0])
            traces.append(plotdat)

        # Set axes
        for plotnum in range(nplots):
            if plotnum == 0:
                self.axes = self.figure.add_subplot(nplots, 1, plotnum+1)
                ax = self.axes
            else:
                self.axes = self.figure.add_subplot(nplots, 1, plotnum+1, sharex=ax)
                setp(self.axes.get_xticklabels(), visible=False)

            # Create OPlots for each subfigure
            try:
                for data in traces[plotnum]:
                    label = data.desc
                    if label == "":
                        label = data.name + " (" + data.units + ")"
                    label += " " + data.source

                    time = data.time
                    if time is None:
                        if len(data.dim) != 1:
                            raise ValueError("Cannot plot '"+data.name+"' as it has too many dimensions")
                        time = data.dim[0].data

                    self.axes.plot(time, data.data, label=label)

                # Y label from last plot
                ylabel = data.desc
                if ylabel == "":
                    ylabel = data.label
                    if data.units != "":
                        ylabel += " ("+data.units+")"
                self.axes.set_ylabel(ylabel)

                if plotnum == 0:
                    self.axes.set_xlabel(traces[0][0].dim[traces[0][0].order].label)

                self.axes.legend()

            except TypeError:
                # Traces[plotnum] not iterable
                time = traces[plotnum].time.data
                xlabel = traces[plotnum].dim[traces[plotnum.order]].label
                if time is None:
                    if len(traces[plotnum].dim) != 1:
                        raise ValueError("Cannot plot '"+traces[plotnum].label+"' as it has too many dimensions")
                    time = traces[plotnum].dim[0].data
                    xlabel = traces[plotnum].dim[0].label

                data = traces[plotnum].data
                if size > 10000:
                    fac = int(len(time) / 10000)
                    time = time[::fac]
                    data = data[::fac]
                    print("Warning: too many samples (%d). Down-sampling to %d points" % (size, len(time)))

                self.axes.plot(time, data)

                ylabel = traces[plotnum].desc
                if ylabel == "":
                    ylabel = traces[plotnum].label
                    if traces[plotnum].units != "":
                        ylabel += " ("+traces[plotnum].units+")"
                self.axes.set_ylabel(ylabel)
                if plotnum == 0:
                    self.axes.set_xlabel(xlabel)

        self.figure.subplots_adjust(left=0.08, right=0.98, top=0.95, bottom=0.07, hspace=0.001)
        self.canvas.draw()

    def zplot(self, *args):
        """
        Plot traces and a zoomed in section on subplots
        """

        def connect_bbox(bbox1, bbox2,
                         loc1a, loc2a, loc1b, loc2b,
                         prop_lines, prop_patches=None):
            """
            Connect two bounding boxes
            """
            if prop_patches is None:
                prop_patches = prop_lines.copy()
                prop_patches["alpha"] = prop_patches.get("alpha", 1)*0.2

            c1 = BboxConnector(bbox1, bbox2, loc1=loc1a, loc2=loc2a, **prop_lines)
            c1.set_clip_on(False)
            c2 = BboxConnector(bbox1, bbox2, loc1=loc1b, loc2=loc2b, **prop_lines)
            c2.set_clip_on(False)

            bbox_patch1 = BboxPatch(bbox1, **prop_patches)
            bbox_patch2 = BboxPatch(bbox2, **prop_patches)

            patch = BboxConnectorPatch(bbox1, bbox2,
                                       loc1a=loc1a, loc2a=loc2a, loc1b=loc1b, loc2b=loc2b,
                                       **prop_patches)
            patch.set_clip_on(False)

            return c1, c2, bbox_patch1, bbox_patch2, patch

        def zoom_effect(ax1, ax2, **kwargs):
            """
            ax1 : the main axes
            ax1 : the zoomed axes

            Connect ax1 and ax2.  The xmin & xmax will be taken from the
            ax1.viewLim.
            """

            tt = ax1.transScale + (ax1.transLimits + ax2.transAxes)
            trans = blended_transform_factory(ax2.transData, tt)

            mybbox1 = ax1.bbox
            mybbox2 = TransformedBbox(ax1.viewLim, trans)

            prop_patches = kwargs.copy()
            prop_patches["ec"] = "none"
            prop_patches["alpha"] = 0.2

            c1, c2, bbox_patch1, bbox_patch2, patch = \
                connect_bbox(mybbox1, mybbox2,
                             loc1a=3, loc2a=2, loc1b=4, loc2b=1,
                             prop_lines=kwargs, prop_patches=prop_patches)

            ax1.add_patch(bbox_patch1)
            ax2.add_patch(bbox_patch2)
            ax2.add_patch(c1)
            ax2.add_patch(c2)
            ax2.add_patch(patch)

            return c1, c2, bbox_patch1, bbox_patch2, patch

        # if len(args) == 1:
        #     self.plot(*args)

        self._clean_axes()

        self.ax_top = self.figure.add_subplot(2, 1, 1)
        self.ax_bottom = self.figure.add_subplot(2, 1, 2)
        self.axes = [self.ax_top, self.ax_bottom]

        zoom_effect(self.ax_top, self.ax_bottom)
        self.start_cid = self.ax_bottom.figure.canvas.mpl_connect("button_press_event",
                                                                  self.start_zoom_region)
        self.end_cid = self.ax_bottom.figure.canvas.mpl_connect("button_release_event",
                                                                self.end_zoom_region)
        self.callback_id = [self.start_cid, self.end_cid]

        for tracenum, trace in enumerate(args):
            try:
                for data in trace:
                    label = data.desc
                    if label == "":
                        label = data.name + " (" + data.units + ") " + data.source

                    time = data.time
                    if time is None:
                        if len(data.dim) != 1:
                            print(data.dim)
                            raise ValueError("Cannot plot '"+data.label+"' as it has too many dimensions")
                        time = data.dim[0].data

                    self.ax_top.plot(time, data.data, label=label)
                    self.ax_bottom.plot(time, data.data, label=label)
                if tracenum == 0:
                    self.ax_bottom.set_xlabel(trace[0].dim[trace[0].order].label)
                    ylabel = data.desc
                    if ylabel == "":
                        ylabel = data.label
                        if data.units != "":
                            ylabel += " ("+data.units+") "
                    self.ax_bottom.set_ylabel(ylabel)
                    self.ax_top.set_ylabel(ylabel)
                self.ax_bottom.legend()
            except TypeError:
                # Trace not iterable
                time = trace.time.data
                xlabel = trace.dim[trace.order].label
                if time is None:
                    if len(trace.dim) != 1:
                        raise ValueError("Cannot plot '"+trace.label+"' as it has too many dimensions")
                    time = trace.dim[0].data
                    xlabel = trace.dim[0].label
                self.ax_bottom.set_xlabel(xlabel)

                data = trace.data
                # Check array size
                size = len(time)
                if size > 10000:
                    fac = int(len(time) / 10000)
                    time = time[::fac]
                    data = data[::fac]
                    print("Warning: too many samples (%d). Down-sampling to %d points" % (size, len(time)))

                self.ax_top.plot(time, data)
                self.ax_bottom.plot(time, data)

                ylabel = trace.desc
                if ylabel == "":
                    ylabel = trace.label
                    if trace.units != "":
                        ylabel += " ("+trace.units+") "
                self.ax_top.set_ylabel(ylabel)
                self.ax_bottom.set_ylabel(ylabel)

        self.figure.subplots_adjust(left=0.07, right=0.98, top=0.95, bottom=0.08)
        self.canvas.draw()

    def start_zoom_region(self, event):
        """
        Callback for starting the zoomed region
        """
        if event.inaxes != self.ax_bottom:
            return
        _, right = self.ax_top.get_xlim()
        self.ax_top.set_xlim(event.xdata, right)
        self.ax_bottom.figure.canvas.draw()
        self.drag_cid = self.ax_bottom.figure.canvas.mpl_connect("motion_notify_event",
                                                                 self.middle_zoom_region)

    def middle_zoom_region(self, event):
        """
        Callback for dragging the zoomed region
        """
        if event.inaxes != self.ax_bottom:
            return
        left, _ = self.ax_top.get_xlim()
        self.ax_top.set_xlim(left, event.xdata)
        self.ax_bottom.figure.canvas.draw()

    def end_zoom_region(self, event):
        """
        Callback for ending the zoomed region
        """
        self.figure.canvas.mpl_disconnect(self.drag_cid)

    def contour(self, item, levels=10):
        """
        Contour plot of item

        item - Item to plot
        levels - Number of contours to plot [10]
        """

        self._clean_axes()
        self.axes = self.figure.add_subplot(111)
        self.figure.subplots_adjust(left=0.07, right=0.98, top=0.95, bottom=0.08)

        if not (len(item.dim) == 2 or
                len(item.data) == 2 or
                len(item.data.shape) == 2):
            raise ValueError("Data must be two dimensional")
            return

        xdim = item.dim[item.order]
        ydim = item.dim[item.order+1]

        xaxis = xdim.data
        yaxis = ydim.data
        zaxis = item.data
        self.axes.contour(xaxis, yaxis, zaxis, levels)

        xlabel = xdim.label
        if xlabel == "":
            xlabel = xdim.name
            if xdim.units != "":
                xlabel += " ({})".format(xdim.units)

        self.axes.set_xlabel(xlabel)

        ylabel = ydim.label
        if ylabel == "":
            ylabel = ydim.name
            if ydim.units != "":
                ylabel += " ({})".format(ydim.units)

        self.axes.set_ylabel(ylabel)

        title = item.name
        if title == "":
            title = item.title
        self.axes.set_title(title)

        self.canvas.draw()

    def contourf(self, item, levels=10):
        """
        Filled contour plot of item

        item - Item to plot
        levels - Number of contours to plot [10]
        """

        self._clean_axes()
        self.axes = self.figure.add_subplot(111)
        self.figure.subplots_adjust(left=0.07, right=0.98, top=0.95, bottom=0.08)

        if not (len(item.dim) == 2 or
                len(item.data) == 2 or
                len(item.data.shape) == 2):
            raise ValueError("Data must be two dimensional")
            return

        xdim = item.dim[item.order]
        ydim = item.dim[item.order+1]

        xaxis = xdim.data
        yaxis = ydim.data
        zaxis = item.data
        self.axes.contourf(xaxis, yaxis, zaxis, levels)

        xlabel = xdim.label
        if xlabel == "":
            xlabel = xdim.name
            if xdim.units != "":
                xlabel += " ({})".format(xdim.units)

        self.axes.set_xlabel(xlabel)

        ylabel = ydim.label
        if ylabel == "":
            ylabel = ydim.name
            if ydim.units != "":
                ylabel += " ({})".format(ydim.units)

        self.axes.set_ylabel(ylabel)

        title = item.name
        if title == "":
            title = item.title
        self.axes.set_title(title)

        self.canvas.draw()

    def clearFig(self):
        """
        Reset the plot widget
        """
        self._clean_axes()
        self.axes = self.figure.add_subplot(111)
        self.figure.subplots_adjust(left=0.08, right=0.98, top=0.95, bottom=0.07)
        self.canvas.draw()
